@@ -78,6 +78,13 @@ static const gchar introspection_xml[] =
 	"      <arg type='h' name='input_stream' direction='in' />"
 	"      <arg type='aaa{ss}' name='result' direction='out' />"
 	"    </method>"
+	"    <method name='Deserialize'>"
+	"      <arg type='h' name='input_stream' direction='in' />"
+	"      <arg type='i' name='flags' direction='in' />"
+	"      <arg type='i' name='format' direction='in' />"
+	"      <arg type='s' name='default_graph' direction='in' />"
+	"      <arg type='a{sv}' name='arguments' direction='in' />"
+	"    </method>"
 	"    <signal name='GraphUpdated'>"
 	"      <arg type='sa{ii}' name='updates' />"
 	"    </signal>"
@@ -367,7 +374,8 @@ write_cursor (QueryRequest          *request,
 		glong cur_offset = -1;
 
 		if (!g_data_output_stream_put_int32 (request->data_stream, n_columns,
-		                                     NULL, &inner_error))
+		                                     request->cancellable,
+		                                     &inner_error))
 			break;
 
 		for (i = 0; i < n_columns; i++) {
@@ -375,7 +383,8 @@ write_cursor (QueryRequest          *request,
 
 			if (!g_data_output_stream_put_int32 (request->data_stream,
 			                                     tracker_sparql_cursor_get_value_type (cursor, i),
-			                                     NULL, &inner_error))
+			                                     request->cancellable,
+			                                     &inner_error))
 				goto out;
 
 			values[i] = tracker_sparql_cursor_get_string (cursor, i, &len);
@@ -386,17 +395,21 @@ write_cursor (QueryRequest          *request,
 
 		for (i = 0; i < n_columns; i++) {
 			if (!g_data_output_stream_put_int32 (request->data_stream,
-			                                     offsets[i], NULL, &inner_error))
+			                                     offsets[i],
+			                                     request->cancellable,
+			                                     &inner_error))
 				goto out;
 		}
 
 		for (i = 0; i < n_columns; i++) {
 			if (!g_data_output_stream_put_string (request->data_stream,
 			                                      values[i] ? values[i] : "",
-			                                      NULL, &inner_error))
+			                                      request->cancellable,
+			                                      &inner_error))
 				goto out;
 
-			if (!g_data_output_stream_put_byte (request->data_stream, 0, NULL,
+			if (!g_data_output_stream_put_byte (request->data_stream, 0,
+			                                    request->cancellable,
 			                                    &inner_error))
 				goto out;
 		}
@@ -424,7 +437,8 @@ handle_cursor_reply (GTask        *task,
 	TrackerSparqlCursor *cursor = source_object;
 	QueryRequest *request = task_data;
 	const gchar **variable_names = NULL;
-	GError *write_error = NULL;
+	GError *error = NULL;
+	gboolean retval;
 	gint i, n_columns;
 
 	n_columns = tracker_sparql_cursor_get_n_columns (cursor);
@@ -434,14 +448,15 @@ handle_cursor_reply (GTask        *task,
 
 	g_dbus_method_invocation_return_value (request->invocation, g_variant_new ("(^as)", variable_names));
 
-	if (!write_cursor (request, cursor, &write_error))
-		g_warning ("Endpoint failed to fully write cursor: %s\n", write_error->message);
+	retval = write_cursor (request, cursor, &error);
 	g_free (variable_names);
-	g_clear_error (&write_error);
 
 	tracker_sparql_cursor_close (cursor);
 
-	g_task_return_boolean (task, TRUE);
+	if (error)
+		g_task_return_error (task, error);
+	else
+		g_task_return_boolean (task, retval);
 }
 
 static void
@@ -685,6 +700,25 @@ read_update_blank_cb (GObject      *object,
 	                                              request);
 }
 
+static void
+deserialize_cb (GObject      *object,
+                GAsyncResult *res,
+                gpointer      user_data)
+{
+	UpdateRequest *request = user_data;
+	GError *error = NULL;
+
+	if (!tracker_sparql_connection_deserialize_finish (TRACKER_SPARQL_CONNECTION (object),
+	                                                   res, &error)) {
+		g_dbus_method_invocation_return_gerror (request->invocation, error);
+		update_request_free (request);
+		return;
+	}
+
+	g_dbus_method_invocation_return_value (request->invocation, NULL);
+	update_request_free (request);
+}
+
 static TrackerSparqlStatement *
 create_statement (TrackerSparqlConnection  *conn,
                   const gchar              *query,
@@ -926,6 +960,49 @@ endpoint_dbus_iface_method_call (GDBusConnection       *connection,
 			request = update_request_new (endpoint_dbus, invocation, FALSE, fd);
 			update_request_read_next (request, read_update_blank_cb);
 		}
+	} else if (g_strcmp0 (method_name, "Deserialize") == 0) {
+		TrackerDeserializeFlags flags;
+		TrackerRdfFormat format;
+		gchar *graph;
+
+		if (tracker_endpoint_dbus_forbid_operation (endpoint_dbus,
+		                                            invocation,
+		                                            TRACKER_OPERATION_TYPE_UPDATE)) {
+			g_dbus_method_invocation_return_error (invocation,
+			                                       G_DBUS_ERROR,
+			                                       G_DBUS_ERROR_ACCESS_DENIED,
+			                                       "Operation not allowed");
+			return;
+		}
+
+		g_variant_get (parameters, "(hiisa{sv})", &handle, &flags, &format, &graph, &arguments);
+
+		if (fd_list)
+			fd = g_unix_fd_list_get (fd_list, handle, &error);
+
+		if (fd < 0) {
+			g_dbus_method_invocation_return_error (invocation,
+			                                       G_DBUS_ERROR,
+			                                       G_DBUS_ERROR_INVALID_ARGS,
+			                                       "Did not get a file descriptor");
+		} else {
+			TrackerSparqlConnection *conn;
+			UpdateRequest *request;
+
+			request = update_request_new (endpoint_dbus, invocation, FALSE, fd);
+			conn = tracker_endpoint_get_sparql_connection (TRACKER_ENDPOINT (request->endpoint));
+
+			tracker_sparql_connection_deserialize_async (conn,
+			                                             flags,
+			                                             format,
+			                                             graph && *graph ? graph : NULL,
+			                                             G_INPUT_STREAM (request->input_stream),
+			                                             request->endpoint->cancellable,
+			                                             deserialize_cb,
+			                                             request);
+		}
+
+		g_free (graph);
 	} else {
 		g_dbus_method_invocation_return_error (invocation,
 		                                       G_DBUS_ERROR,
